import websockets, asyncio
from .errors import ServerError, ServerClosedError, ServerUnreachableError
from json import dumps, loads


class MultiplayerClient:
    def __init__(self, msg_handler, ip="127.0.0.1", port=1300, auth_handler=None):
        self.ip = ip
        self.port = port
        self.ws = None
        self.id = None
        self._msg_handler = msg_handler
        self._auth_handler = auth_handler

    async def _run(self, proxy):
        try:
            async with websockets.connect(f"ws://{self.ip}:{self.port}") as websocket:
                if self._auth_handler:
                    await self._auth_handler(websocket)

                msg = await websocket.recv()
                uri = loads(msg)["content"]
                if not uri.startswith("ws://"):
                    print("Invalid URI")
                    await websocket.close()
                    quit()

                await websocket.close()

            async with websockets.connect(uri) as websocket:
                self.ws = websocket
                await proxy(websocket)

        except OSError:
            raise ServerUnreachableError(self.ip, self.port)

    async def msg_handler(self):
        try:
            async for msg_json in self.ws:
                msg = loads(msg_json)
                if msg["type"] == "error":
                    raise ServerError(msg["content"])

                elif msg["type"] == "id":
                    self.id = msg["content"]
                    print(self.id)

                await self._msg_handler(msg, self.ws)
        
        except websockets.exceptions.ConnectionClosedError:
            raise ServerClosedError()

    def run(self, proxy):
        asyncio.run(self._run(proxy))



import websockets
import asyncio


async def proxy(websocket):
    print("Connected")
    await websocket.send("login")
    name = await websocket.recv()
    password = await websocket.recv()
    with open("users.txt", "r") as f:
        users = [user for user in f.read().split("\n")]

    if f"{name} {password}" in users:
        await websocket.send("success")

    else:
        await websocket.send("failure")



async def main():
    async with websockets.serve(proxy, "localhost", 3000):
        await asyncio.Future()


if __name__ == "__main__":
    asyncio.run(main())



class PortInUseError(Exception):
    def __init__(self, port):
        self.message = f"Port {port} is already in use."
        super().__init__(self.message)

class ServerUnreachableError(Exception):
    def __init__(self, ip, port):
        self.message = f"The server at {ip}:{port} could not be reached."
        super().__init__(self.message)


class AuthServerOffline(Exception):
    def __init__(self):
        self.message = "The provided authentication server could not be reached."
        super().__init__(self.message)


class ServerClosedError(Exception):
    def __init__(self):
        self.message = "The server forcibly closed the connection. This is likely due to the server encountering an error." # NOQA
        super().__init__(self.message)

class ServerError(Exception):
    def __init__(self, msg):
        self.message = (f"The server encountered an error and had to close the connection. It sent this error: {msg}")
        super().__init__(self.message)



from .TCPserver import TCPMultiplayerServer
from .client import MultiplayerClient
print("This library is currently undergoing a major rewrite.")
print("Not all features will work as intended.")
print("pymultiplayer by iamdeedz")
print("https://www.github.com/iamdeedz/pymultiplayer")



from pymultiplayer.client import MultiplayerClient
from pymultiplayer.errors import ServerError
from json import dumps, loads


async def auth_handler(websocket):
    await websocket.recv()
    name = input("Enter name: ")
    password = input("Enter password: ")
    await websocket.send(name)
    await websocket.send(password)
    response_json = await websocket.recv()
    response = loads(response_json)
    if response["type"] == "error":
        raise ServerError(response["content"])

    elif response["content"] == "success":
        print("Authenticated.")
        
    else:
        print("Authentication failed.")
        await websocket.close()


async def msg_handler(msg_json, websocket):
    msg = loads(msg_json)
    print("server sent:", msg["content"])


async def proxy(websocket):
    print("Connected")
    txt = input("enter message: ")
    msg = {"type": "message", "content": txt}
    await websocket.send(dumps(msg))
    await client.msg_handler(websocket)


if __name__ == "__main__":
    client = MultiplayerClient(msg_handler, auth_handler=auth_handler)
    client.run(proxy)



from pymultiplayer.TCPserver import TCPMultiplayerServer
from json import dumps
import websockets


async def msg_handler(msg, client):
    print(f"Client with id {client.id}:", msg["content"])
    print("Sending back:", msg["content"])
    await client.ws.send(dumps(msg))


async def auth_func(websocket):
    print("Authenticating...")
    await websocket.send("login")
    name = await websocket.recv()
    password = await websocket.recv()
    async with websockets.connect("ws://localhost:3000") as ws:
        await ws.recv()
        await ws.send(name)
        await ws.send(password)
        response = await ws.recv()
        await ws.close()

    if response == "success":
        print("Authenticated.")
        msg = {"type": "authentication", "content": "success"}
        await websocket.send(dumps(msg))

    else:
        print("Authentication failed.")
        msg = {"type": "authentication", "content": "failure"}
        await websocket.send(dumps(msg))
        await websocket.close()



if __name__ == "__main__":
    server = TCPMultiplayerServer(msg_handler, auth_func=auth_func)
    server.run()



from pymultiplayer import MultiplayerClient
from json import dumps, loads
from player import Player
import pygame as p

id = None
self = None
other_players = []


async def main():
    for event in p.event.get():
        if event.type == p.QUIT:
            global running
            running = False


async def msg_handler(msg_json, client):
    if msg["type"] == "client_joined":
        global other_players
        other_players.append(Player(msg["content"]))


async def proxy(websocket):
    global self
    self = Player(id)
    while running:
        await client.msg_handler()
        await main()

    await websocket.close()


if __name__ == "__main__":
    p.init()
    p.display.set_caption("Multiplayer Test")
    screen = p.display.set_mode((500, 500))
    clock = p.time.Clock()
    running = True

    client = MultiplayerClient(msg_handler)
    client.run(proxy)



import websockets, asyncio
from threading import Thread
from .errors import PortInUseError, AuthServerOffline
from json import dumps


class InitialServer:
    def __init__(self, ip="127.0.0.1", port=1300, auth_func=None):
        self.ip = ip
        self.port = port
        self._auth_func = auth_func

    async def _start(self):
        try:
            async with websockets.serve(self.new_client, self.ip, self.port):
                await asyncio.Future()
        except OSError:
            raise PortInUseError(self.port)

    async def new_client(self, websocket):
        if self._auth_func:
            try:
                await self._auth_func(websocket)
            except OSError:
                msg = {"type": "error", "content": "Server encountered an OSError during the authorization process."}
                await websocket.send(dumps(msg))
                await websocket.close()
                raise AuthServerOffline()

        msg = {"type": "uri", "content": f"ws://{self.ip}:{self.port + 1}"}
        await websocket.send(dumps(msg))
        await websocket.close()

    def start(self):
        asyncio.run(self._start())



from pymultiplayer import TCPMultiplayerServer
from json import dumps, loads


async def msg_handler(msg, client):
    if msg["type"] == "greeting":
        await client.ws.send(dumps())


if __name__ == "__main__":
    server = TCPMultiplayerServer(msg_handler)
    server.run()



import websockets, asyncio
from ._ws_client import _Client
from .initial_server import InitialServer
from threading import Thread
from json import dumps, loads


class TCPMultiplayerServer:
    def __init__(self, msg_handler, ip="127.0.0.1", port=1300, auth_func=None):
        self.ip = ip
        self.port = port
        self.msg_handler = msg_handler
        self.clients = set()
        self.last_id = 0
        self.initial_server = InitialServer(self.ip, self.port, auth_func)
        Thread(target=self.initial_server.start).start()

    def broadcast(self, msg):
        client_websockets = [client.ws for client in self.clients]
        websockets.broadcast(client_websockets, msg)

    async def _run(self):
        try:
            async with websockets.serve(self.proxy, self.ip, self.port + 1):
                await asyncio.Future()
        except OSError:
            raise PortInUseError(self.port)

    async def proxy(self, websocket, path):
        new_client = _Client(websocket, self.last_id + 1)
        self.last_id += 1

        try:
            self.clients.add(new_client)
            await websocket.send(dumps({"type": "id", "content": new_client.id}))
            for client in self.clients:
                if client.ws == websocket:
                    continue
                msg = {"type": "client_joined", "content": new_client.id}
                await client.ws.send(dumps(msg))
            print(f"Client with id {self.last_id} connected")
            async for msg_json in websocket:
                client = [client for client in self.clients if client.ws == websocket][0]
                msg = loads(msg_json)
                await self.msg_handler(msg, client)

        finally:
            self.clients.remove(new_client)
            msg = {"type": "client_left", "content": client.id}
            self.broadcast(dumps(msg))


    def run(self):
        asyncio.run(self._run())



from websocket_server import WebsocketServer
from re import match
from .errors import *
from multiprocessing import Process
import websocket
import socket
import rel


class UDPMultiplayerServer:
    def __init__(self, ip="127.0.0.1", port=1300):
        self.ip = ip
        self.port = port
        self.server = None
        self.msg_handler = None
        self.clients = []
        self.initial_server = InitialServer(ip, port)
        self.start()

    def start(self):
        try:
            self.server = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

            self.server.bind((self.ip, self.port + 1))

        except OSError:
            self.initial_server.server.shutdown_gracefully()
            raise PortInUseError(self.port)

        Process(target=self.msg_handler).start()

        print(f"Hosting server at IP address {socket.gethostbyname(socket.gethostname())} on port {self.port}. "
              "Share this with friends to play together!")

    def close(self):

        self.server.close()

    def send(self, msg, client):
        self.server.sendto(str.encode(msg), client)

    def send_to_all(self, msg):
        for client in self.clients:
            self.server.sendto(str.encode(msg), client)

    def msg_received(self, msg, client):
        pass

    def new_client(self, client, server):
        pass

    def client_left(self, client, server):
        pass

    def set_msg_received_func(self, func):
        self.msg_received = func  # NOQA

    def set_new_client_func(self, func):
        self.new_client = func  # NOQA

    def set_client_left_func(self, func):
        self.client_left = func  # NOQA

    def msg_handler(self):
        while True:
            msg, client = self.server.recvfrom(1024)

            if match("greeting", msg.decode()):
                self.clients.append(client)
                self.new_client(client, self.server)
                continue

            if match("goodbye", msg.decode()):
                self.clients.remove(client)
                self.client_left(client, self.server)
                continue

            self.msg_received(msg, client)


class TCPMultiplayerServer:
    def __init__(self, ip="127.0.0.1", port=1300):
        self.ip = ip
        self.port = port
        self.server = None
        self.clients = []
        self.initial_server = InitialServer(ip, port)
        self.start()

    def start(self):
        try:
            self.server = WebsocketServer(host=self.ip, port=self.port + 1)
        except OSError:
            self.initial_server.server.shutdown_gracefully()
            raise PortInUseError(self.port)

        self.server.set_fn_new_client(self.new_client)

        print(f"Hosting server at IP address {socket.gethostbyname(socket.gethostname())} on port {self.port}. "
              "Share this with friends to play together!")

    def run_forever(self):
        self.server.run_forever()

    def send(self, msg, client):
        self.server.send(msg, client)

    def send_to_all(self, msg):
        self.server.send_to_all(msg)

    def new_client(self, client, server):
        self.clients.append(client)

    def msg_received(self):
        pass

    def set_msg_received_func(self, func):
        self.server.set_fn_message_received(func)

    def set_new_client_func(self, func):
        self.server.set_fn_new_client(func)

    def set_client_left_func(self, func):
        self.server.set_fn_client_left(func)


# ---------------------------------------------------------------------------------------- #


class InitialServer:
    def __init__(self, ip="127.0.0.1", port=1300):
        self.ip = ip
        self.port = port
        self.server = None
        self.start()

    def start(self):
        try:
            self.server = WebsocketServer(host=self.ip, port=self.port)
        except OSError:
            raise PortInUseError(self.port)
        self.server.set_fn_new_client(self.new_client)
        self.server.run_forever(threaded=True)

    def send_to_main_server(self):
        self.server.send_message_to_all(f"ws://{self.ip}:{self.port + 1}")

    def new_client(self, client, server):
        self.send_to_main_server()

    def set_new_client_func(self, func):
        self.server.set_fn_new_client(func)


# ---------------------------------------------------------------------------------------- #


class MultiplayerClient:
    def __init__(self, ip="127.0.0.1", port=1300, tick_func=None, args=()):

        if tick_func is None:
            raise NoTickFunctionError()

        else:
            self.tick_func = tick_func
            self.tick_func_args = (self,) + args

        self.ip = ip
        self.port = port
        self.server = None
        self.protocol = None

    def connect(self):
        try:
            ws = websocket.WebSocket()
            ws.connect(f"ws://{self.ip}:{self.port}")
            main_server = ws.recv()
            ws.close()

        except OSError:
            raise ServerRefusedError(self.ip, self.port)

        if match("ws://", main_server):

            self.protocol = "TCP"

            # Connect to TCP server
            self.server = websocket.WebSocketApp(main_server)

            self.server.run_forever(dispatcher=rel)

            rel.signal(2, self.disconnect)
            rel.timeout(1, self.tick_func, *self.tick_func_args)

            try:
                rel.dispatch()

            except ConnectionResetError:
                raise ServerClosedError()

        else:

            # Connect to UDP server
            self.server = socket.socket(family=socket.AF_INET, type=socket.SOCK_DGRAM)
            Process(target=self.msg_handler).start()
            self.server.sendto(str.encode("greeting"), (self.ip, self.port + 1))

    def disconnect(self):
        if self.protocol == "TCP":
            self.server.send("goodbye")

        else:
            self.server.sendto(str.encode("goodbye"), (self.ip, self.port + 1))

        rel.abort()
        quit()

    def send(self, msg):
        if self.protocol == "TCP":
            self.server.send(msg)
        else:
            self.server.sendto(str.encode(msg), (self.ip, self.port + 1))

    def msg_received(self, msg):
        pass

    def on_error(self, ws, error):
        pass

    def on_close(self):
        pass

    def set_msg_received_func(self, func):
        if self.protocol == "TCP":
            self.server.on_message = func

        else:
            self.msg_received = func  # NOQA

    def set_on_error_func(self, func):
        if self.protocol == "TCP":
            self.server.on_error = func

        else:
            print("The host is using a UDP based server which does not support error handling.")

    def set_on_close_func(self, func):
        if self.protocol == "TCP":
            self.server.on_close = func

        else:
            self.on_close = func  # NOQA

    def msg_handler(self):
        while True:
            msg, _ = self.server.recvfrom(1024)

            if match("close", msg.decode()):
                self.on_close()

            self.msg_received(msg.decode())
